import qs from 'querystring'

import sql from '@databases/sql'
import axios from 'axios'
import ConnectPG from 'connect-pg-simple'
import fastify, { FastifyPluginAsync } from 'fastify'
import fCookie from 'fastify-cookie'
import rateLimit from 'fastify-rate-limit'
import fSession from 'fastify-session'
import fastifySwagger from 'fastify-swagger'

import { db, isDev } from '../shared'
import { allRouter } from './all'
import { userRouter } from './user'

const apiRouter: FastifyPluginAsync = async (f) => {
  f.register(fCookie)
  f.register(fSession, {
    secret: process.env['SECRET']!,
    // @ts-ignore
    store: new (ConnectPG(fSession))(
      process.env['DATABASE_URL']
        ? {
            conString: process.env['DATABASE_URL'],
          }
        : {
            conObject: {
              user: process.env['PGUSER'],
              password: process.env['PGPASSWORD'],
              database: process.env['PGDATABASE'],
            },
          }
    ),
  })

  f.register(rateLimit, {
    max: 10,
    timeWindow: '1 second',
    allowList: (req) => {
      if (req.routerPath === '/api/quiz/getSrsLevel') {
        return true
      }

      return false
    },
  })

  f.register(fastifySwagger, {
    openapi: {
      security: [
        {
          BearerAuth: [],
        },
      ],
      components: {
        securitySchemes: {
          BearerAuth: {
            type: 'http',
            scheme: 'bearer',
          },
        },
      },
    },
    exposeRoute: isDev,
    routePrefix: '/doc',
  })

  f.addHook('preHandler', async (req) => {
    const { body, log } = req

    if (body && typeof body === 'object' && body.constructor === Object) {
      log.info({ body }, 'parsed body')
    }
  })

  f.addHook('preHandler', async (req) => {
    if (['/api/doc', '/api/settings'].some((s) => req.url.startsWith(s))) {
      return
    }

    if (req.session['userId']) {
      return
    }

    const [, apiKey] =
      /^Bearer (.+)$/.exec(req.headers.authorization || '') || []

    if (!apiKey) {
      throw { statusCode: 401, message: 'no apiKey in header' }
    }

    const userData = await axios
      .create({
        baseURL: 'https://api.wanikani.com/v2/',
        headers: {
          Authorization: `Bearer ${apiKey}`,
        },
        validateStatus: function () {
          return true
        },
      })
      .get('/user')
      .then(
        (r) =>
          r.data as {
            data: {
              id: string
              username: string
              level: number
              subscription: {
                active: boolean
                type: string
              }
            }
          }
      )

    const userId = userData.data.id

    await db.query(sql`
    INSERT INTO "user" ("id", "identifier", "level")
    VALUES (${userId}, ${userData.data.username}, ${userData.data.level})
    ON CONFLICT ("id") DO UPDATE SET
      "level" = EXCLUDED."level"
    `)

    req.session['userId'] = userId
  })

  f.register(allRouter, {
    prefix: '/all',
  })

  f.register(userRouter, {
    prefix: '/user',
  })
}

export default apiRouter

if (require.main === module) {
  async function main() {
    const app = fastify({
      logger: {
        prettyPrint: isDev,
        serializers: {
          req(req) {
            if (!req.url) {
              return { method: req.method }
            }

            const [url = '', q] = req.url.split(/\?(.+)$/)
            const query = q ? qs.parse(q) : undefined

            return { method: req.method, url, query }
          },
        },
      },
    })

    app.register(apiRouter, {
      prefix: '/api',
    })

    return app.listen(process.env['PORT']!, '0.0.0.0')
  }
  main()
}
